#include                <sys/stat.h>
#include                <linux/types.h>
#include                <errno.h>

#include                <string.h>
#include                <fcntl.h>
#include                <stdio.h>
#include                <stdlib.h>
#include                <unistd.h>
#include                <sys/syscall.h>
#include                <dirent.h>
#include                <utime.h>
#include                <sys/mount.h>
#include                "tap.h"
#include                "libFireBird.h"

#define PROGRAM_NAME    "SmartBackup"
#define VERSION         "V1.0"

TAP_ID                  (0x8E0A426E);
TAP_PROGRAM_NAME        (PROGRAM_NAME" "VERSION);
TAP_AUTHOR_NAME         ("FireBird");
TAP_DESCRIPTION         ("Backup everything from ProgramFiles");
TAP_ETCINFO             (__DATE__);

#define STACKTRACE      FALSE
#define ROOTDIR         "/ProgramFiles/Settings/"PROGRAM_NAME
#define INIFILE         PROGRAM_NAME".ini"
#define SBDBHDR         "SBDB"
#define SBDBVERSION     1
#define SBDBNAME        "SBDB.db"

typedef struct
{
  __dev_t               st_dev;
  long int              st_pad1[2];
  __ino64_t             st_ino;
  __mode_t              st_mode;
  __nlink_t             st_nlink;
  __uid_t               st_uid;
  __gid_t               st_gid;
  __dev_t               st_rdev;
  long int              st_pad2[2];
  __off64_t             st_size;
  __time_t              st_atime;
  long int              __reserved0;
  __time_t              st_mtime;
  long int              __reserved1;
  __time_t              st_ctime;
  long int              __reserved2;
  __blksize_t           st_blksize;
  long int              st_pad4;
  __blkcnt64_t          st_blocks;
  long int              st_pad5[14];
} tstat64;

extern int lstat64(__const char *__restrict __file, tstat64 *__restrict __buf) __THROW;

typedef enum
{
  ST_Init,
  ST_WaitMount,
  ST_RecursiveScan,
  ST_CopyFiles,
  ST_Cleanup,
  ST_Exit
} tState;

typedef struct
{
  int                   FileNameIndex;
  int                   PathIndex;
  dword                 PathCRC;
  __off64_t             FileSize;
  __time_t              CTime;
  __time_t              MTime;
  byte                  Hash[16];
}tDBEntry;

typedef struct
{
  char                  DBType[4];
  dword                 Version;
  dword                 UpdateTime;
  int                   DBSize;
  int                   DBNrEntries;
  tDBEntry             *DBEntry;
  int                   TextDBSize;
  int                   TextDBEnd;
  char                 *TextDB;
} tSBDB;

tState                  State = ST_Init;
char                    LogString[512];
char                    INIDestDevice[20];
char                    INIDestPath[128];
char                    INIMountCommand[256];
dword                   INILastBackup;
dword                   LastBackup;
char                    SourceRoot[256];
char                    DestRoot[256];
char                    LastDestRoot[256];
dword                   StateTimeout = 0;
bool                    SBDidNASMount;
tSBDB                   CurrentSBDB, LastSBDB;
ulong64                 TotalSourceFileSize;
int                     TotalSourceFiles, TotalSourceDirs;
int                     TotalLinkedFiles, TotalLinkTries, TotalLinkFailures, TotalCopiedFiles;
ulong64                 TotalLinkedSize, TotalCopiedSize;
int                     TotalCopyFailures;
ulong64                 TotalCopyFailureSize;

void WriteLog(char *s)
{
  static bool                 FirstCall = TRUE;

  CallTraceEnter("WriteLog");

  HDD_TAP_PushDir();
  if(FirstCall)
  {
    HDD_ChangeDir("/ProgramFiles");
    TAP_Hdd_Create("Settings", ATTR_FOLDER);
    HDD_ChangeDir("Settings");
    TAP_Hdd_Create(PROGRAM_NAME, ATTR_FOLDER);
    FirstCall = FALSE;
  }
  HDD_ChangeDir(ROOTDIR);
  StrMkISO(s);
  LogEntryGeneric(PROGRAM_NAME, TRUE, s);
  HDD_TAP_PopDir();

  CallTraceExit(NULL);
}

void SaveINI(void)
{
  CallTraceEnter("SaveINI");

  HDD_TAP_PushDir();
  HDD_ChangeDir(ROOTDIR);
  INIOpenFile (INIFILE, PROGRAM_NAME);

  INISetString("DestDevice", INIDestDevice);
  INISetString("DestPath", INIDestPath);
  INISetString("MountCommand", INIMountCommand);
  INISetHexWord("LastBackup", INILastBackup);

  INISaveFile (INIFILE, INILOCATION_AtCurrentDir, NULL);
  INICloseFile();
  HDD_TAP_PopDir();

  CallTraceExit(NULL);
}

void LoadINI(void)
{
  CallTraceEnter("LoadINI");

  INILOCATION           INILocation;

  HDD_TAP_PushDir();
  HDD_ChangeDir(ROOTDIR);
  INILocation = INIOpenFile(INIFILE, NULL);
  if(INILocation == INILOCATION_NewFile)
  {
    WriteLog("SmartBackup.ini not found!");
    INICloseFile();
  }
  else
  {
    INIGetString("DestDevice", INIDestDevice, "", sizeof(INIDestDevice));
    INIGetString("DestPath", INIDestPath, "", sizeof(INIDestPath));
    INIGetString("MountCommand", INIMountCommand, "", sizeof(INIMountCommand));
    INILastBackup = INIGetHexWord("LastBackup", 0, 0, 0xffff);
    INICloseFile();
  }
  HDD_TAP_PopDir();

  CallTraceExit(NULL);
}

void INIUpdateBackupDate(void)
{
  CallTraceEnter("INIUpdateBackupDate");

  INILastBackup = LastBackup >> 16;
  HDD_TAP_PushDir();
  HDD_ChangeDir(ROOTDIR);
  INIOpenFile (INIFILE, PROGRAM_NAME);

  INISetHexWord("LastBackup", INILastBackup);

  INISaveFile (INIFILE, INILOCATION_AtCurrentDir, NULL);
  INICloseFile();
  HDD_TAP_PopDir();

  CallTraceExit(NULL);
}

void FileSizeToString(ulong64 FileSize, char *SizeString)
{
  CallTraceEnter("FileSizeToString");

  dword                 s;

  if(SizeString)
  {
    if((FileSize >> 20) != 0)
    {
      //Hier geht es in MB weiter
      s = (dword)(FileSize >> 10);

      if((s >> 20) == 0)
      {
        TAP_SPrint(SizeString, "%.1f MB", (float)s / 1024);
      }
      else if((s >> 30) == 0)
      {
        TAP_SPrint(SizeString, "%.1f GB", (float)(s >> 10) / 1024);
      }
      else
      {
        TAP_SPrint(SizeString, "%.1f TB", (float)(s >> 20) / 1024);
      }
    }
    else
    {
      // < 1MB
      s = FileSize;
      if((s >> 10) == 0)
      {
        TAP_SPrint(SizeString, "%d B", s);
      }
      else
      {
        TAP_SPrint(SizeString, "%.1f kB", (float)s / 1024);
      }
    }
  }

  CallTraceExit(NULL);
}

bool FindExtBackupDisk(char *DestRoot)
{
  char                  c;
  byte                  d;
  char                  FileName[80];
  tstat64               Stat;

  CallTraceEnter("FindExtBackupDisk");

  DestRoot[0] = '\0';

  for(c = 'b'; c <= 'z'; c++)
  {
    //Real original disks
    TAP_SPrint(FileName, "/mnt/sd%c%s/", c, INIDestPath);
    if(lstat64(FileName, &Stat) == 0)
    {
      strcpy(DestRoot, FileName);
      CallTraceExit(NULL);
      return TRUE;
    }

    for(d = 1; d < 9; d++)
    {
      TAP_SPrint(FileName, "/mnt/sd%c%d%s/", c, d, INIDestPath);
      if(lstat64(FileName, &Stat) == 0)
      {
        strcpy(DestRoot, FileName);
        CallTraceExit(NULL);
        return TRUE;
      }
    }
  }

  CallTraceExit(NULL);
  return FALSE;
}

bool FindNASDisk(char *DestRoot)
{
  tstat64               Stat;
  bool                  ret;

  CallTraceEnter("FindNASDisk");

  ret = FALSE;
  DestRoot[0] = '\0';

  if((lstat64(INIDestPath, &Stat) == 0) && Stat.st_dev != 0x0802)
  {
    strcpy(DestRoot, INIDestPath);
    ret = TRUE;
  }

  CallTraceExit(NULL);
  return ret;
}

void FreeSBDB(tSBDB *SDBD)
{
  CallTraceEnter("FreeSBDB");

  if(SDBD)
  {
    if(SDBD->DBEntry) TAP_MemFree(SDBD->DBEntry);
    if(SDBD->TextDB) TAP_MemFree(SDBD->TextDB);
    memset(SDBD, 0, sizeof(tSBDB));

    //Damit der erste Eintrag immer ein leerer String bleibt
    SDBD->TextDBEnd = 1;
  }

  CallTraceExit(NULL);
}

bool LoadSBDB(char *AbsFileName, tSBDB *SDBD)
{
  CallTraceEnter("LoadSBDB");

  bool                  ret;
  int                   f, x;

  FreeSBDB(SDBD);
  ret = FALSE;
  if(AbsFileName && SDBD)
  {
    f = open(AbsFileName, O_RDONLY, 0600);
    if(f >= 0)
    {
      //Header lesen
      if(read(f, SDBD, sizeof(tSBDB)) == sizeof(tSBDB))
      {
        //Gltiger Header?
        if((memcmp(SDBD->DBType, SBDBHDR, 4) == 0) && (SDBD->Version == SBDBVERSION))
        {
          SDBD->DBSize = SDBD->DBNrEntries;
          SDBD->TextDBSize = SDBD->TextDBEnd;

          if(SDBD->DBSize != 0)
          {
            x = SDBD->DBNrEntries * sizeof(tDBEntry);
            SDBD->DBEntry = TAP_MemAlloc(x);

            //Speicherplatz und Einlesen der Datenbankeintrge ok?
            if(SDBD->DBEntry && (read(f, SDBD->DBEntry, x) == x))
            {
              //Und jetzt noch fr die Texteintrge
              SDBD->TextDB = TAP_MemAlloc(SDBD->TextDBSize);
              if(SDBD->TextDB && (read(f, SDBD->TextDB, SDBD->TextDBSize) == SDBD->TextDBSize))
              {
                //Jetzt wurde die DB erfolgreich eingelesen
                ret = TRUE;
              }
            }
          }
          else
          {
            //Sollte es keinen einzigen Eintrag geben, ist das Laden trotzdem OK
            ret = TRUE;
          }
        }
      }
      close(f);
    }
  }

  CallTraceExit(NULL);
  return ret;
}

bool SaveSBDB(char *AbsFileName, tSBDB *SDBD)
{
  CallTraceEnter("SaveSBDB");

  bool                  ret;
  int                   f;

  ret = FALSE;

  if(AbsFileName && SDBD)
  {
    f = creat(AbsFileName, 0777);
    if(f >= 0)
    {
      //Header lesen
      memcpy(SDBD->DBType, SBDBHDR, 4);
      SDBD->Version = SBDBVERSION;
      SDBD->UpdateTime = LastBackup;
      write(f, SDBD, sizeof(tSBDB));
      write(f, SDBD->DBEntry, SDBD->DBNrEntries * sizeof(tDBEntry));
      write(f, SDBD->TextDB, SDBD->TextDBEnd);
      close(f);
    }
    else
      WriteLog("Failed to save the SmartBackup database");
  }

  CallTraceExit(NULL);
  return ret;
}

int AddTextToSBDB(char *Text, tSBDB *SDBD)
{
  CallTraceEnter("AddTextToSBDB");

  int                   ret;
  int                   x;
  char                 *TempText, *p;

  ret = -1;
  if(Text && SDBD)
  {
    if(SDBD->TextDBSize > 0)
    {
      //Eine Schleife durch die komplette TextDB laufen und nachsehen, ob der Text schon existiert
      x = 0;
      p = &SDBD->TextDB[0];
      while(x < SDBD->TextDBEnd)
      {
        if(strcmp(Text, p) == 0)
        {
          //Den Text gibt es schon
          ret = x;
          break;
        }
        x += (strlen(p) + 1);
        p += (strlen(p) + 1);
      }
    }

    if(ret == -1)
    {
      //Nicht gefunden, hinten anhngen
      //Muss die Text-Datenbank zuerst erweitert werden?
      if((int)(SDBD->TextDBEnd + strlen(Text) + 2) >= SDBD->TextDBSize)
      {
        //um 1kB erweitern
        x = (SDBD->TextDBEnd + strlen(Text) + 2 + 1024);
        TempText = TAP_MemAlloc(x);
        if(TempText)
        {
          memset(TempText, 0, x);
          if(SDBD->TextDBSize > 0)
          {
            memcpy(TempText, SDBD->TextDB, SDBD->TextDBEnd);
            TAP_MemFree(SDBD->TextDB);
          }
          SDBD->TextDB = TempText;
          SDBD->TextDBSize = x;
        }
        else
        {
          TAP_SPrint(LogString, "Failed to increase the SDBD text to %d characters!", x);
          WriteLog(LogString);

          State = ST_Exit;
          CallTraceExit(NULL);
          return -1;
        }
      }

      ret = SDBD->TextDBEnd;
      strcpy(&SDBD->TextDB[SDBD->TextDBEnd], Text);
      SDBD->TextDBEnd += (strlen(Text) + 1);
    }
  }

  CallTraceExit(NULL);
  return ret;
}

tDBEntry *AddToSBDB(char *AbsFileName, tSBDB *SDBD)
{
  CallTraceEnter("AddToSBDB");

  tDBEntry           *pSBDBEntry;
  tstat64               Stat;
  char                 *p, c, *RelPath;

  pSBDBEntry = NULL;
  if(SDBD && (lstat64(AbsFileName, &Stat) == 0))
  {
    //Muss die File-Datenbank erweitert werden?
    if(SDBD->DBNrEntries >= SDBD->DBSize)
    {
      //um 100 Eintrge erweitern
      int                x;
      tDBEntry        *TempSDBDEntry;

      x = (SDBD->DBSize + 100) * sizeof(tDBEntry);
      TempSDBDEntry = TAP_MemAlloc(x);
      if(TempSDBDEntry)
      {
        memset(TempSDBDEntry, 0, x);
        if(SDBD->DBSize > 0)
        {
          memcpy(TempSDBDEntry, SDBD->DBEntry, SDBD->DBNrEntries * sizeof(tDBEntry));
          TAP_MemFree(SDBD->DBEntry);
        }

        SDBD->DBEntry = TempSDBDEntry;
        SDBD->DBSize += 100;
      }
      else
      {
        TAP_SPrint(LogString, "Failed to increase the SDBD to %d entries!", SDBD->DBSize + 100);
        WriteLog(LogString);

        State = ST_Exit;
        CallTraceExit(NULL);
        return NULL;
      }
    }
    pSBDBEntry = &SDBD->DBEntry[SDBD->DBNrEntries];
    pSBDBEntry->CTime = Stat.st_ctime;
    pSBDBEntry->MTime = Stat.st_mtime;
    pSBDBEntry->FileSize = Stat.st_size;
    MD5AbsFile(AbsFileName, pSBDBEntry->Hash);

    //RelPath zeigt hinter /mnt/hd/ProgramFiles und somit auf den relativen Pfad
    RelPath = &AbsFileName[strlen(SourceRoot)];
    pSBDBEntry->PathCRC = CRC32(0, RelPath, strlen(RelPath));

    //Absoluten Pfad in Pfad und Dateinamen aufspalten
    p = strrchr(RelPath, '/');
    if(p)
    {
      c = p[1];
      p[1] = '\0';
      pSBDBEntry->PathIndex = AddTextToSBDB(RelPath, SDBD);

      p[1] = c;
      pSBDBEntry->FileNameIndex = AddTextToSBDB(&p[1], SDBD);
    }
    else
    {
      pSBDBEntry->PathIndex = 0;
      pSBDBEntry->FileNameIndex = AddTextToSBDB(RelPath, SDBD);
    }

    SDBD->DBNrEntries++;
  }

  CallTraceExit(NULL);
  return pSBDBEntry;
}

bool RecursiveScan(void)
{
  CallTraceEnter("AddToSBDB");

  typedef enum
  {
    SST_Init,
    SST_NextDir,
    SST_NextFile,
    SST_Done
  }tScanState;

  static tScanState         ScanState = SST_Init;
  static int                i, NrFiles;
  static TYPE_FolderEntry   FolderEntry;
  static tStringDB          StringDB;
  static int                MaxLoop;
  bool                      ret;
  char                      s[512];
  tDBEntry               *SBDBEntry;

  ret = FALSE;

  switch(ScanState)
  {
    case SST_Init:
    {
      WriteLog("Performing recursive scan");
      FreeSBDB(&CurrentSBDB);
      TotalSourceFileSize = 0;
      TotalSourceFiles = 0;
      TotalSourceDirs = 0;
      MaxLoop = 0;

      HDD_TAP_PushDir();

      //Perform a full recursive search for the file
      StringDBInit(&StringDB, 1024);
      StringDBAdd(&StringDB, "/ProgramFiles");
      ScanState = SST_NextDir;
      break;
    }

    case SST_NextDir:
    {
      if(!StringDBEOF(&StringDB))
      {
        if(MaxLoop++ > 1000)
        {
          WriteLog("Recursive scan seems to be in an endless loop!");
          HDD_TAP_PopDir();

          CallTraceExit(NULL);
          return TRUE;
        }

        HDD_ChangeDir(StringDBGet(&StringDB, StringDBCurrent(&StringDB)));
        TotalSourceDirs++;
        NrFiles = TAP_Hdd_FindFirst(&FolderEntry, "");
        i = 0;
        ScanState = SST_NextFile;
      }
      else
      {
        ScanState = SST_Done;
      }
      break;
    }

    case SST_NextFile:
    {
      if(i < NrFiles)
      {
        if(FolderEntry.attr == ATTR_FOLDER)
        {
          TAP_SPrint(s, "%s/%s", StringDBGet(&StringDB, StringDBCurrent(&StringDB)), FolderEntry.name);
          StringDBAdd(&StringDB, s);
        }

        if(FolderEntry.attr == ATTR_NORMAL)
        {
          TAP_SPrint(s, "%s%s/%s", TAPFSROOT, StringDBGet(&StringDB, StringDBCurrent(&StringDB)), FolderEntry.name);
          SBDBEntry = AddToSBDB(s, &CurrentSBDB);
          if(SBDBEntry)
          {
            TotalSourceFiles++;
            TotalSourceFileSize += SBDBEntry->FileSize;
          }
        }

        TAP_Hdd_FindNext(&FolderEntry);
        i++;
      }
      else
      {
        StringDBNext(&StringDB);
        ScanState = SST_NextDir;
      }
      break;
    }

    case SST_Done:
    {
      FileSizeToString(TotalSourceFileSize, s);
      TAP_SPrint(LogString, "Scan result: %d folders, %d files, %s", TotalSourceDirs, TotalSourceFiles, s);
      WriteLog(LogString);

      StringDBDestroy(&StringDB);
      HDD_TAP_PopDir();
      ret = TRUE;
      break;
    }
  }

  CallTraceExit(NULL);
  return ret;
}

void FindLastBackup(tSBDB *SBDB)
{
  CallTraceEnter("FindLastBackup");

  DIR                  *dirp;
  struct dirent        *dp;
  tstat64               Stat;
  char                  FullPath[512];
  tSBDB                 TempSBDB;
  word                  year;
  byte                  month, day, weekday;

  //Alle Zielverzeichnisse durchsuchen und die neueste Datenbank laden = letztes erfolgreiches Backup
  FreeSBDB(SBDB);
  if(SBDB)
  {
    dirp = opendir(DestRoot);
    if(dirp)
    {
      while((dp = readdir(dirp)) != NULL)
      {
        TAP_SPrint(FullPath, "%s%s", DestRoot, dp->d_name);

        if((lstat64(FullPath, &Stat) == 0) && S_ISDIR(Stat.st_mode))
        {
          //Verzeichnis: prfen, ob es darin eine SBDB gibt
          strcat(FullPath, "/" SBDBNAME);
          if(LoadSBDB(FullPath, &TempSBDB))
          {
            //Wenn das Laden erfolgreich war und die TempSDDB neuer als die per Parameter bergebene SBDB ist, dann umkopieren
            if(TempSBDB.UpdateTime > SBDB->UpdateTime)
            {
              FreeSBDB(SBDB);
              memcpy(SBDB, &TempSBDB, sizeof(tSBDB));
              TAP_SPrint(LastDestRoot, "%s%s/", DestRoot, dp->d_name);
            }
            else
              FreeSBDB(&TempSBDB);
          }
        }
      }
      closedir(dirp);
    }
  }

  if(SBDB->UpdateTime == 0)
  {
    strcpy(LogString, "No previous backup found, this might be the first one.");
  }
  else
  {
    TAP_ExtractMjd(SBDB->UpdateTime >> 16, &year, &month, &day, &weekday);
    TAP_SPrint(LogString, "Last backup found: %4.4d-%2.2d-%2.2d", year, month, day);
  }
  WriteLog(LogString);

  CallTraceExit(NULL);
}

bool CopyFile(char *AbsSourceName, char *AbsDestName)
{
  CallTraceEnter("CopyFile");

  int                   fIn, fOut;
  byte                  Buffer[4096];
  int                   BytesRead;
  bool                  ret;
  tstat64               Stat;
  struct utimbuf        times;

  ret = FALSE;
  lstat64(AbsSourceName, &Stat);
  fIn = open(AbsSourceName, O_RDONLY, 0600);
  if(fIn >= 0)
  {
    fOut = creat(AbsDestName, 0777);
    if(fOut >= 0)
    {
      do
      {
        BytesRead = read(fIn, Buffer, 4096);
        if(BytesRead > 0) write(fOut, Buffer, BytesRead);
      } while(BytesRead > 0);
      ret = TRUE;

      close(fOut);

      times.actime = Stat.st_mtime;
      times.modtime = Stat.st_mtime;
      utime(AbsDestName, &times);
    }
    else
    {
      TAP_SPrint(LogString, "Failed to open destination '%s'", AbsSourceName);
      WriteLog(LogString);
    }
    close(fIn);
  }
  else
  {
    TAP_SPrint(LogString, "Failed to open source '%s'", AbsSourceName);
    WriteLog(LogString);
  }

  CallTraceExit(NULL);
  return ret;
}

bool DoBackup(tSBDB *CurrentSBDB, tSBDB *LastSBDB)
{
  CallTraceEnter("DoBackup");

  typedef enum
  {
    DBK_Init,
    DBK_NextFile,
    DBK_Done
  }tDoBackupState;

  static tDoBackupState     DoBackupState = DBK_Init;
  static int                i, j;
  static char               LastDir[512];
  bool                      ret;
  char                      s[512];
  tDBEntry                 *pCurrentDBEntry, *pLastDBEntry;
  bool                      Found;
  char                     *pPath, *pFileName;
  char                      SourceFile[512], DestFile[512], LastDestFile[512];

  ret = FALSE;

  switch(DoBackupState)
  {
    case DBK_Init:
    {
      i = 0;
      TotalCopiedFiles = 0;
      TotalCopiedSize = 0;
      TotalLinkedFiles = 0;
      TotalLinkedSize = 0;
      TotalLinkTries = 0;
      TotalLinkFailures = 0;
      TotalCopyFailures = 0;
      TotalCopyFailureSize = 0;

      DoBackupState = DBK_NextFile;
      break;
    }

    case DBK_NextFile:
    {
      if(i < CurrentSBDB->DBNrEntries)
      {
        pCurrentDBEntry = &CurrentSBDB->DBEntry[i];

        pPath = &CurrentSBDB->TextDB[pCurrentDBEntry->PathIndex];
        pFileName = &CurrentSBDB->TextDB[pCurrentDBEntry->FileNameIndex];

        //Zielverzeichnis anlegen
        TAP_SPrint(DestFile, "%s%s", DestRoot, pPath);
        TAP_SPrint(s, "mkdir -p \"%s\"\n", DestFile);
        if(strcmp(s, LastDir))
        {
          system(s);
          strcpy(LastDir, s);
        }

        //Vollstndigen Quell- und Zielnamen zusammenbauen
        TAP_SPrint(SourceFile, "%s%s%s", SourceRoot, pPath, pFileName);
        TAP_SPrint(DestFile, "%s%s%s", DestRoot, pPath, pFileName);

        //Nachsehen, ob es die Datei schon im letzten Backup gab
        pLastDBEntry = &LastSBDB->DBEntry[0];
        Found = FALSE;
        for(j = 0; j < LastSBDB->DBNrEntries; j++)
        {
          if(pCurrentDBEntry->PathCRC == pLastDBEntry->PathCRC)
          {
            Found = TRUE;

            //Die Datei gab es beim letzten Backup. Ist sie auch identisch?
            if(memcmp(pCurrentDBEntry->Hash, pLastDBEntry->Hash, 16) == 0)
            {
              //Ja, mittels Hard-Link verknpfen
              TAP_SPrint(LastDestFile, "%s%s%s", LastDestRoot, pPath, pFileName);
              TotalLinkTries++;
              if(syscall(SYS_link, LastDestFile, DestFile) == 0)
              {
                //Hard link scheint geklappt zu haben, Statistik updaten
                TotalLinkedFiles++;
                TotalLinkedSize += pCurrentDBEntry->FileSize;
              }
              else
              {
                //Link is fehlgeschlagen, kopieren
                if(CopyFile(SourceFile, DestFile))
                {
                  TotalCopiedFiles++;
                  TotalCopiedSize += pCurrentDBEntry->FileSize;
                  TotalLinkFailures++;
                }
                else
                {
                  TotalCopyFailures++;
                  TotalCopyFailureSize += pCurrentDBEntry->FileSize;
                }
              }
            }
            else
            {
              //Nein, kopieren
              if(CopyFile(SourceFile, DestFile))
              {
                TotalCopiedFiles++;
                TotalCopiedSize += pCurrentDBEntry->FileSize;
              }
              else
              {
                TotalCopyFailures++;
                TotalCopyFailureSize += pCurrentDBEntry->FileSize;
              }
            }
          }
          pLastDBEntry++;
        }

        if(!Found)
        {
          //Es handelt sich um eine neue Datei, also kopieren
          if(CopyFile(SourceFile, DestFile))
          {
            TotalCopiedFiles++;
            TotalCopiedSize += pCurrentDBEntry->FileSize;
          }
          else
          {
            TotalCopyFailures++;
            TotalCopyFailureSize += pCurrentDBEntry->FileSize;
          }
        }

        i++;
      }
      else
      {
        DoBackupState = DBK_Done;
      }
      break;
    }

    case DBK_Done:
    {
      //Statistik ausgeben
      FileSizeToString(TotalCopiedSize, s);
      TAP_SPrint(LogString, "Copied: %d files, %s", TotalCopiedFiles, s);
      WriteLog(LogString);

      FileSizeToString(TotalLinkedSize, s);
      TAP_SPrint(LogString, "DeDupe: %d files, %s", TotalLinkedFiles, s);
      if(TotalLinkFailures > 0)
      {
        TAP_SPrint(s, " (%d of %d hard links failed)", TotalLinkFailures, TotalLinkTries);
        strcat(LogString, s);
      }
      WriteLog(LogString);

      if(TotalCopyFailures)
      {
        FileSizeToString(TotalCopyFailureSize, s);
        TAP_SPrint(LogString, "Failed: %d files, %s", TotalCopyFailures, s);
        WriteLog(LogString);
      }

      ret = TRUE;
      break;
    }
  }

  CallTraceExit(NULL);
  return ret;
}

void MountNAS(void)
{
  CallTraceEnter("MountNAS");

  char *p;

  if(INIMountCommand[0])
  {
    SBDidNASMount = TRUE;
    TAP_SPrint(LogString, "Launching mount command '%s'", INIMountCommand);
    WriteLog(LogString);

    //Mount-Verzeichnis erstellen, letztes Leerzeichen im Mount-Befehl suchen
    p = strrchr(INIMountCommand, ' ');
    if(p)
    {
      TAP_SPrint(LogString, "mkdir -p %s", &p[1]);
      system(LogString);
    }

    TAP_SPrint(LogString, "%s &", INIMountCommand);
    system(LogString);
  }

  CallTraceExit(NULL);
}

void UnmountNAS(void)
{
  CallTraceEnter("UnmountNAS");

  char *p;

  //Letztes Leerzeichen im Mount-Befehl suchen
  p = strrchr(INIMountCommand, ' ');
  if(p)
  {
    umount(&p[1]);
    rmdir(&p[1]);
  }

  CallTraceExit(NULL);
}

void CleanupBackups(void)
{
  CallTraceEnter("CleanupBackups");

  #define LISTSIZE        100

  typedef struct
  {
    word                  BackupDate;
    word                  year;
    byte                  month, day, weekday;
  }tBackupList;

  DIR                  *dirp;
  struct dirent        *dp;
  tstat64               Stat;
  char                  FullPath[512];
  tSBDB                 TempSBDB;
  tBackupList           BackupList[LISTSIZE], *pBackupList;
  int                   NrBackups;

  //Zielverzeichniss durchsuchen und alle Backups einlesen
  FreeSBDB(&TempSBDB);
  NrBackups = 0;
  dirp = opendir(DestRoot);
  if(dirp)
  {
    while((dp = readdir(dirp)) != NULL)
    {
      TAP_SPrint(FullPath, "%s%s", DestRoot, dp->d_name);

      if((lstat64(FullPath, &Stat) == 0) && S_ISDIR(Stat.st_mode))
      {
        //Verzeichnis: prfen, ob es darin eine SBDB gibt
        strcat(FullPath, "/" SBDBNAME);
        if(LoadSBDB(FullPath, &TempSBDB))
        {
          //Wenn das Laden erfolgreich war, die Daten in ein Array eintragen
          if(NrBackups < LISTSIZE)
          {
            pBackupList = &BackupList[NrBackups];
            NrBackups++;

            pBackupList->BackupDate = TempSBDB.UpdateTime;
            TAP_ExtractMjd(pBackupList->BackupDate, &pBackupList->year, &pBackupList->month, &pBackupList->day, &pBackupList->weekday);
          }
          FreeSBDB(&TempSBDB);
        }
      }
    }
    closedir(dirp);
  }

  CallTraceExit(NULL);
}

dword TAP_EventHandler(word event, dword param1, dword param2)
{
  static bool           DoNotReenter = FALSE;
  static dword          NextCheck = 0;

  if(DoNotReenter) return param1;
  DoNotReenter = TRUE;

  CallTraceEnter("TAP_EventHandler");

  (void) param2;

  if(event == EVT_IDLE)
  {
    switch(State)
    {
      case ST_Init:
      {
        word            mjd;
        byte            h, m, s;
        tstat64         statbuf;

        //Prfen, ob heute schon ein Backup gemacht wurde
        SBDidNASMount = FALSE;
        LoadINI();
        TAP_GetTime(&mjd, &h, &m, &s);
        if(INILastBackup == mjd)
        {
          WriteLog("Today's backup has been done already.");
          State = ST_Exit;
          break;
        }
        LastBackup = Now(NULL);

        //Ziel prfen
        strcpy(SourceRoot, "/mnt/hd/ProgramFiles");
        State = ST_RecursiveScan;
        LowerCase(INIDestDevice);
        switch(INIDestDevice[0])
        {
          case 'i':
          {
            if(!INIDestPath[0] || (INIDestPath[strlen(INIDestPath) - 1] != '/')) strcat(INIDestPath, "/");
            strcpy(DestRoot, INIDestPath);
            TAP_SPrint(LogString, "Destination: internal disk @ '%s'", DestRoot);
            WriteLog(LogString);

            TAP_SPrint(LogString, "mkdir -p %s", DestRoot);
            system(LogString);

            if(lstat64(DestRoot, &statbuf) != 0)
            {
              WriteLog("Destination path does not exist!");
              State = ST_Exit;
              break;
            }

            break;
          }

          case 'e':
          {
            if(!FindExtBackupDisk(DestRoot))
              WriteLog("Destination: external disk not found, waiting up to 3 minutes");

            StateTimeout = TAP_GetTick() + 30000;
            NextCheck = 0;
            State = ST_WaitMount;

            break;
          }

          case 'n':
          {
            if(!FindNASDisk(DestRoot))
            {
              WriteLog("Destination: NAS not mounted, waiting up to 3 minutes");
              MountNAS();
            }

            StateTimeout = TAP_GetTick() + 30000;
            NextCheck = 0;
            State = ST_WaitMount;

            break;
          }

          default:
          {
            WriteLog("No valid destination, check SmartBackup.ini!");
            State = ST_Exit;
            break;
          }
        }

        break;
      }

      case ST_WaitMount:
      {
        if(TAP_GetTick() > NextCheck)
        {
          switch(INIDestDevice[0])
          {
            //Hier kommen wir nicht her, da auf die interne Disk nie gewartet wird
            case 'i': break;

            case 'e':
            {
              if(FindExtBackupDisk(DestRoot))
              {
                TAP_SPrint(LogString, "Destination: external disk @ '%s'", DestRoot);
                WriteLog(LogString);
                State = ST_RecursiveScan;
              }
            }

            case 'n':
            {
              if(FindNASDisk(DestRoot))
              {
                TAP_SPrint(LogString, "Destination: NAS @ '%s'", DestRoot);
                WriteLog(LogString);
                State = ST_RecursiveScan;
              }
              break;
            }
          }

          if(TAP_GetTick() > StateTimeout)
          {
            WriteLog("Destination: mount wait timeout!");
            State = ST_Exit;
          }

          NextCheck = TAP_GetTick() + 500;
        }
        break;
      }

      case ST_RecursiveScan:
      {
        word            year;
        byte            month, day, weekday;

        if(RecursiveScan())
        {
          FindLastBackup(&LastSBDB);

          //Heutiges Datum an Zielpfad anhngen
          TAP_ExtractMjd(LastBackup >> 16, &year, &month, &day, &weekday);
          TAP_SPrint(&DestRoot[strlen(DestRoot)], "%4.4d-%2.2d-%2.2d/", year, month, day);

          //Verzeichnis erzeugen. shell mkdir verwenden, damit alle Unterverzeichnisse miterzeugt werden
          TAP_SPrint(LogString, "mkdir -p \"%s\"", DestRoot);
          system(LogString);

          State = ST_CopyFiles;
        }
        break;
      }

      case ST_CopyFiles:
      {
        if(DoBackup(&CurrentSBDB, &LastSBDB))
        {
          //Datenbank speichern
          TAP_SPrint(LogString, "%s%s", DestRoot, SBDBNAME);
          SaveSBDB(LogString, &CurrentSBDB);
          INIUpdateBackupDate();
          State = ST_Exit;
        }
        break;
      }

      case ST_Cleanup:
      {
        CleanupBackups();
        break;
      }

      case ST_Exit:
      {
        //Falls SB den NAS-Mount ausgefhrt hat, dann auch wieder unmounten
        if(SBDidNASMount) UnmountNAS();

        WriteLog("Exiting");
        TAP_Exit();
        break;
      }
    }
  }

  if(event == EVT_STOP) //srp abgeschalten oder tap beenden
  {
    TAP_Exit();
  }

  CallTraceExit(NULL);
  DoNotReenter = FALSE;
  return param1;
}

int TAP_Main(void)
{
  CallTraceInit();
  CallTraceEnable(STACKTRACE);
  CallTraceEnter("TAP_Main");

  TAP_SPrint(LogString, PROGRAM_NAME " " VERSION);
  WriteLog(LogString);

  CallTraceExit(NULL);
  return 1;
}
